@@ -38,11 +38,8 @@
 
 #include <string.h>
 
-#define FILE_NAME      "FILE"
-
 #define LOC_DELIMITER   ":"
 #define LOC_DELIMITER_C ':'
-#define PART_DELIMITER  "-"
 
 typedef struct _GkrLocationVolume {
 	GQuark volume_loc;
@@ -240,7 +237,7 @@ hal_device_removed (LibHalContext *hal_ctx, const char *udi)
 	}
 
 	name = udi_to_location_name (hal_ctx, udi);
-	g_assert (name && name[0]);
+	g_return_if_fail (name && name[0]);
 	
 	if (g_hash_table_lookup (pv->volumes_by_name, name)) {
 		g_message ("removing removable location: %s", name); 
@@ -279,7 +276,7 @@ hal_device_property (LibHalContext *hal_ctx, const char *udi, const char *key,
 		goto done;
 
 	name = udi_to_location_name (hal_ctx, udi);
-	g_assert (name && name[0]);
+	g_return_if_fail (name && name[0]);
 	
 	/* A mount was added? */
 	if (is_mounted &&  !g_hash_table_lookup (pv->volumes_by_name, name)) {
@@ -405,9 +402,13 @@ location_manager_hal_uninit (GkrLocationManager *locmgr)
 	DBusError error;
 
 	if (pv->hal_ctx) {
-		if (!libhal_ctx_shutdown (pv->hal_ctx, &error))
+		dbus_error_init (&error);
+		if (!libhal_ctx_shutdown (pv->hal_ctx, &error)) {
 			g_warning ("failed to shutdown HAL context: %s\n", error.message);
-		else if (!libhal_ctx_free (pv->hal_ctx)) 
+			dbus_error_free (&error);
+		} 
+		
+		if (!libhal_ctx_free (pv->hal_ctx)) 
 			g_warning ("failed to free HAL context");
 		pv->hal_ctx = NULL;
 	}
@@ -441,7 +442,7 @@ gkr_location_manager_init (GkrLocationManager *locmgr)
 	g_return_if_fail (home && home[0]);
 	
 	/* Hidden location relative to file system and home directory */
-	make_hidden_volume (locmgr, &pv->file_volume, FILE_NAME, "/", NULL);
+	make_hidden_volume (locmgr, &pv->file_volume, GKR_LOCATION_NAME_FILE, "/", NULL);
 	make_hidden_volume (locmgr, &pv->home_volume, GKR_LOCATION_NAME_HOME, home, _("Home"));
 
 	
@@ -454,7 +455,6 @@ gkr_location_manager_init (GkrLocationManager *locmgr)
 	if (!local)
 		local = g_build_filename (home, ".gnome2", NULL);
 
-	g_assert (local);
 	gkr_location_manager_register (locmgr, GKR_LOCATION_NAME_LOCAL, local, _("Home"));
 	g_free (local);
 	
@@ -729,7 +729,7 @@ find_closest_name (gpointer unused, GkrLocationVolume *locvol, FindClosestName *
 GQuark
 gkr_location_from_string (const gchar *str)
 {
-	g_assert (str && str[0]);
+	g_return_val_if_fail (str && str[0], 0);
 	
 	/* TODO: Some sort of validation? */
 	
@@ -746,8 +746,8 @@ gkr_location_from_path (const gchar *path)
 	gchar *loc;
 	const gchar *c;
 	
-	g_assert (path && path[0]);
-	g_assert (g_path_is_absolute (path));
+	g_return_val_if_fail (path && path[0], 0);
+	g_return_val_if_fail (g_path_is_absolute (path), 0);
 	
 	/* We don't allow a colon in our paths */
 	for (c = path; *c; ++c) {
@@ -778,10 +778,10 @@ GQuark
 gkr_location_from_child (GQuark parent, const gchar *child)
 {
 	const gchar *c;
-	gchar *path;
+	gchar *path, *p;
 	GQuark loc;
 	
-	g_assert (parent);
+	g_return_val_if_fail (parent, 0);
 	
 	/* We don't allow a colon in our paths */
 	for (c = child; *c; ++c) {
@@ -792,6 +792,16 @@ gkr_location_from_child (GQuark parent, const gchar *child)
 	}
 
 	path = g_build_path (G_DIR_SEPARATOR_S, g_quark_to_string (parent), child, NULL);
+	
+	/* Strip out trailing slashes */
+	p = path + strlen (path);
+	while (p > path + 1) {
+		*(p--) = 0;
+		if (!G_IS_DIR_SEPARATOR (*p))
+			break;
+	}
+	
+	/* Strip out any trailing slashes */
 	loc = g_quark_from_string (path);
 	g_free (path);
 	return loc;
@@ -806,10 +816,10 @@ location_to_volume (GQuark loc, const gchar **remainder)
 	const gchar *bdelim, *sloc;
 	gchar *name;
 	
-	g_assert (loc);
+	g_return_val_if_fail (loc, NULL);
 
 	sloc = g_quark_to_string (loc);
-	g_assert (sloc);
+	g_return_val_if_fail (sloc, NULL);
 
 	bdelim = strchr (sloc, LOC_DELIMITER_C);
 	if (!bdelim) {
@@ -831,6 +841,32 @@ location_to_volume (GQuark loc, const gchar **remainder)
 	return locvol;
 } 
 
+GQuark
+gkr_location_to_parent (GQuark parent)
+{
+	const gchar *del, *sloc, *part;
+	GQuark ret;
+	gchar *str;
+	
+	g_return_val_if_fail (parent, 0);
+
+	sloc = g_quark_to_string (parent);
+	g_return_val_if_fail (sloc, 0);
+
+	del = strchr (sloc, LOC_DELIMITER_C);
+	if (!del) {
+		g_warning ("The '%s' location is invalid", sloc);
+		return 0;
+	}
+	
+	part = strrchr (del + 1, G_DIR_SEPARATOR);
+	str = g_strndup (sloc, (part ? part : del) - sloc);
+	ret = g_quark_from_string (str);
+	g_free (str);
+	
+	return ret;
+} 
+
 gchar* 
 gkr_location_to_path (GQuark loc)
 {
@@ -839,6 +875,8 @@ gkr_location_to_path (GQuark loc)
 	gsize l;
 	gchar *res;
 	
+	g_return_val_if_fail (loc, NULL);
+	
 	locvol = location_to_volume (loc, &path);
 	if (!locvol)
 		return NULL;
@@ -869,6 +907,9 @@ gkr_location_to_display (GQuark loc)
 	gchar *filename;
 	gchar *display;
 	
+	if (!loc)
+		return g_strdup ("");
+	
 	filename = gkr_location_to_path (loc);
 	if (!filename)
 		return g_strdup ("");
@@ -896,6 +937,8 @@ gkr_location_get_volume (GQuark loc)
 {
 	GkrLocationVolume *locvol;
 	
+	g_return_val_if_fail (loc, 0);
+	
 	locvol = location_to_volume (loc, NULL);
 	if (!locvol)
 		return 0;
