@@ -24,6 +24,7 @@
 #include "config.h"
 
 #include "gkr-pk-index.h"
+#include "gkr-pk-places.h"
 
 #include "common/gkr-async.h"
 #include "common/gkr-cleanup.h"
@@ -52,6 +53,7 @@ struct _GkrPkIndex {
 	 GHashTable *path_by_location;
 	 GHashTable *mtime_by_location;
 	 GHashTable *file_by_location;
+	 GHashTable *defaults_by_parent;
 };
 
 struct _GkrPkIndexClass {
@@ -64,6 +66,12 @@ G_DEFINE_TYPE (GkrPkIndex, gkr_pk_index, G_TYPE_OBJECT);
 static GkrPkIndex *index_singleton = NULL; 
 static GQuark no_location = 0;
 
+typedef gboolean (*ReadValueFunc) (GKeyFile *file, const gchar *group, const gchar *field, 
+                                   GError **err, gpointer user_data);
+                                   
+typedef gboolean (*WriteValueFunc) (GKeyFile *file, const gchar *group, const gchar *field, 
+                                    GError **err, gpointer user_data);
+
 /* -----------------------------------------------------------------------------
  * HELPERS
  */
@@ -107,6 +115,56 @@ free_mtime (gpointer v)
 	g_slice_free (time_t, v);
 }
 
+static GQuark* 
+quarks_from_strings (const gchar **strv, gsize *n_quarks)
+{
+	GArray *arr;
+	GQuark quark;
+	
+	arr = g_array_new (TRUE, TRUE, sizeof (GQuark));
+	while (*strv) {
+		quark = g_quark_from_string (*strv);
+		g_array_append_val (arr, quark);
+		++strv;
+	}
+	
+	if (n_quarks)
+		*n_quarks = arr->len;
+	
+	return (GQuark*)g_array_free (arr, FALSE);
+}
+
+static gchar**
+quarks_to_strings (const GQuark* quarks, gsize *n_strings)
+{
+	const gchar *value;
+	GArray *arr;
+	
+	arr = g_array_new (TRUE, TRUE, sizeof (const gchar*));
+	while (*quarks) {
+		value = g_quark_to_string (*quarks);
+		g_array_append_val (arr, value);
+		++quarks;
+	}
+	
+	if (n_strings)
+		*n_strings = arr->len;
+	return (gchar**)g_array_free (arr, FALSE);
+}
+
+static gboolean
+strings_are_equal (const gchar **one, const gchar **two)
+{
+	while (*one && *two) {
+		if (!g_str_equal (*one, *two))
+			return FALSE;
+		++one;
+		++two;
+	}
+	
+	return *one == *two;
+}
+
 static gpointer
 location_to_key (GQuark loc)
 {
@@ -178,6 +236,8 @@ unique_to_group (gkrconstunique uni)
 	gsize n_group, n_unidata;
 	gboolean r;
 	gchar *group;
+	
+	g_return_val_if_fail (uni, NULL);
 		
 	unidata = gkr_unique_get_raw (uni, &n_unidata);
 	g_assert (unidata);
@@ -191,135 +251,232 @@ unique_to_group (gkrconstunique uni)
 }
 
 static gboolean
-get_keyfile_value (GKeyFile *key_file, gkrconstunique uni, 
-                   const gchar *field, GValue *value)
+read_exists_any_value (GKeyFile *file, const gchar *group, const gchar *field,
+                       GError **err, gboolean *value)
 {
-	GError *err = NULL;
-	gchar *group;
-
-	g_assert (key_file);
-	g_assert (uni);
-	g_assert (field);
 	g_assert (value);
+	*value = g_key_file_has_group (file, group);
+	return TRUE;
+}
 
-	/* TODO: Cache this somehow? */
-	group = unique_to_group (uni);
+static gboolean
+read_exists_value (GKeyFile *file, const gchar *group, const gchar *field, 
+                   GError **err, gboolean *value)
+{
+	g_assert (value);
+	g_assert (field);
+	*value = g_key_file_has_key (file, group, field, err);
+	return *err == NULL;	
+}
 
-	switch (G_VALUE_TYPE (value)) {
-	case G_TYPE_BOOLEAN:
-		{
-			gboolean v = g_key_file_get_boolean (key_file, group, field, &err);
-			if (err == NULL)
-				g_value_set_boolean (value, v);
-		}
-		break;
-		
-	case G_TYPE_INT:
-		{
-			gint v = g_key_file_get_integer (key_file, group, field, &err);
-			if (err == NULL)
-				g_value_set_int (value, v);
-		}
-		break;
-		
-	case G_TYPE_STRING:
-		{
-			gchar *v = g_key_file_get_string (key_file, group, field, &err);
-			if (v != NULL) {
-				g_assert (err == NULL);
-				g_value_take_string (value, v);
-			}
-		}
-		break;
-		
-	default:
-		g_assert_not_reached();
-		break;
+static gboolean
+read_boolean_value (GKeyFile *file, const gchar *group, const gchar *field, 
+                    GError **err, gboolean *value)
+{
+	g_assert (value);
+	g_assert (field);
+	*value = g_key_file_get_boolean (file, group, field, err);
+	return *err == NULL;
+}
+
+static gboolean
+read_int_value (GKeyFile *file, const gchar *group, const gchar *field,
+                GError **err, gint *value)
+{
+	g_assert (value);
+	g_assert (field);
+	*value = g_key_file_get_integer (file, group, field, err);
+	return *err == NULL;
+}
+
+static gboolean
+read_string_value (GKeyFile *file, const gchar *group, const gchar *field,
+                   GError **err, gchar **value)
+{
+	g_assert (value);
+	g_assert (field);
+	*value = g_key_file_get_string (file, group, field, err);
+	return *value != NULL;
+}
+
+static gboolean
+read_quarks_value (GKeyFile *file, const gchar *group, const gchar *field,
+                   GError **err, GQuark **value)
+{
+	gchar **vals;
+	
+	g_assert (value);
+	g_assert (field);
+	
+	vals = g_key_file_get_string_list (file, group, field, NULL, err);
+	if (vals != NULL) {
+		g_assert (*err == NULL);
+		*value = quarks_from_strings ((const gchar**)vals, NULL);
+		g_strfreev (vals);
+		return TRUE;
 	}
 	
-	g_free (group);
+	return FALSE;
+}
+	
+static gint
+get_keyfile_value (GKeyFile *key_file, const gchar *group, 
+                   const gchar *field, ReadValueFunc func, gpointer data)
+{
+	GError *err = NULL;
+	
+	g_assert (key_file);
+	g_assert (group);
+	g_assert (func);
+	
+	if ((func) (key_file, group, field, &err, data))
+		return 1;
 	
 	if (err != NULL) {
 		if (err->code != G_KEY_FILE_ERROR_GROUP_NOT_FOUND &&
-		    err->code != G_KEY_FILE_ERROR_KEY_NOT_FOUND)
+		    err->code != G_KEY_FILE_ERROR_KEY_NOT_FOUND) {
 		    	g_warning ("couldn't read field '%s' from index: %s", 
 		    	           field, err->message ? err->message : "");
+			g_error_free (err);
+			return -1;
+		}
+		
 		g_error_free (err);
+	}
+
+	return 0;
+}
+
+static gboolean
+write_clear (GKeyFile *file, const gchar *group, const gchar *field,
+             GError **err, gpointer user_data)
+{
+	if (!g_key_file_has_group (file, group))
 		return FALSE;
+	g_key_file_remove_group (file, group, err);
+	return TRUE; 	
+}
+
+static gboolean 
+write_delete (GKeyFile *file, const gchar *group, const gchar *field, 
+              GError **err, gpointer user_data)
+{
+	g_assert (field);
+	
+	if (g_key_file_has_key (file, group, field, err)) {
+		g_key_file_remove_key (file, group, field, err);
+		return TRUE;
 	}
 	
-	return TRUE;
+	return FALSE;
+}
+
+static gboolean 
+write_boolean_value (GKeyFile *file, const gchar *group, const gchar *field, 
+                     GError **err, gboolean *value)
+{
+	g_assert (value);
+	g_assert (field);
+	
+	if (g_key_file_get_boolean (file, group, field, err) != *value || err != NULL) {
+		g_clear_error (err);
+		g_key_file_set_boolean (file, group, field, *value);
+		return TRUE;
+	}
+	
+	return FALSE;
+}
+
+static gboolean
+write_int_value (GKeyFile *file, const gchar *group, const gchar *field, 
+                 GError **err, gint *value)
+{
+	g_assert (value);
+	g_assert (field);
+	
+	if (g_key_file_get_integer (file, group, field, err) != *value || err != NULL) {
+		g_clear_error (err);
+		g_key_file_set_integer (file, group, field, *value);
+		return TRUE;
+	}
+	
+	return FALSE;
+}
+
+static gboolean
+write_string_value (GKeyFile *file, const gchar *group, const gchar *field, 
+                    GError **err, const gchar **value)
+{
+	gboolean ret = FALSE;
+	gchar *o;
+	
+	g_assert (value);
+	g_assert (*value);
+	g_assert (field);
+	
+	o = g_key_file_get_value (file, group, field, NULL);
+
+	if (!o || !g_str_equal (o, *value)) {
+		g_key_file_set_string (file, group, field, *value);
+		ret = TRUE;
+	}
+			
+	g_free (o);
+	return ret;
+}
+
+static gboolean
+write_quarks_value (GKeyFile *file, const gchar *group, const gchar *field, 
+                    GError **err, GQuark **value)
+{
+	GQuark *quarks;
+	gsize n_strings;
+	gchar **strings, **o;
+	gboolean ret = FALSE;
+	
+	g_assert (value);
+	g_assert (*value);
+	g_assert (field);
+	
+	quarks = *value;
+	strings = quarks_to_strings (quarks, &n_strings);
+	o = g_key_file_get_string_list (file, group, field, NULL, NULL);
+		
+	if (!o || !strings_are_equal ((const gchar**)strings, (const gchar**)o)) {
+		g_key_file_set_string_list (file, group, field, (const gchar**)strings, n_strings);
+		ret = TRUE;
+	}
+		
+	g_strfreev (o);
+	g_free (strings);
+	return ret;
 }
 
 static void
 set_keyfile_value (GKeyFile *key_file, gkrconstunique uni, 
-                   const gchar *field, const GValue *value,
-                   gboolean *updated)
+                   const gchar *field, WriteValueFunc func, 
+                   gpointer data, gboolean *updated)
 {
 	GError *err = NULL;
 	gchar *group;
 	
 	g_assert (key_file);
 	g_assert (uni);
-	g_assert (field);
-	g_assert (value);
+	g_assert (func);
 	g_assert (updated);
 	
-	*updated = FALSE;
-	
 	/* TODO: Cache this somehow? */
 	group = unique_to_group (uni);
+	g_return_if_fail (group);
 	
-	switch (G_VALUE_TYPE (value)) {
-	case G_TYPE_POINTER:
-		/* GValue can't be set to NULL, so for us empty pointer means delete */
-		g_assert (g_value_get_pointer (value) == NULL);
-		if (g_key_file_has_key (key_file, group, field, NULL)) {
-			g_key_file_remove_key (key_file, group, field, NULL);
-			*updated = TRUE;
-		}
-		break;
-		
-	case G_TYPE_BOOLEAN:
-		{
-			gboolean v = g_value_get_boolean (value);
-			if (g_key_file_get_boolean (key_file, group, field, &err) != v || err != NULL) {
-				g_key_file_set_boolean (key_file, group, field, v);
-				*updated = TRUE;
-			}
-		}
-		break;
-		
-	case G_TYPE_INT:
-		{
-			gint v = g_value_get_int (value);
-			if (g_key_file_get_integer (key_file, group, field, &err) != v || err != NULL) {
-				g_key_file_set_integer (key_file, group, field, v);
-				*updated = TRUE;
-			}
-		}
-		break;
-	case G_TYPE_STRING:
-		{
-			const gchar *v = g_value_get_string (value);
-			gchar *o = g_key_file_get_value (key_file, group, field, &err);
-			
-			g_assert (v != NULL);
-			if (!o || !g_str_equal (o, v)) {
-				g_key_file_set_string (key_file, group, field, v);
-				*updated = TRUE;
-			}
-			
-			g_free (o);
-		}
-		break;
-	default:
-		g_assert_not_reached();
-		break;
+	*updated = (func) (key_file, group, field, &err, data);
+
+	if (err) {
+	    	g_warning ("couldn't write field '%s' to index: %s", 
+	    	           field, err->message ? err->message : "");
+		g_error_free (err);
 	}
 	
-	if (err)
-		g_error_free (err);
 	g_free (group);
 }	
 
@@ -432,15 +589,94 @@ load_index_key_file (GkrPkIndex *index, GQuark loc, int fd, gboolean force)
 	return key_file;
 }
 
+static const gchar*
+find_parent_defaults (GQuark parent)
+{
+	const GkrPkPlace *place;
+	const gchar *defaults = NULL;
+	GSList *volumes, *l;
+	GQuark loc;
+	guint i;
+	
+	for (i = 0; i < G_N_ELEMENTS (gkr_pk_places); ++i) {
+		place = &(gkr_pk_places[i]);
+		
+		/* With a specific volume */
+		if (place->volume) {
+			loc = gkr_location_from_string (place->volume);
+			loc = gkr_location_from_child (loc, place->directory);
+			if (loc == parent)
+				defaults = place->defaults;
+				
+		/* With any volume */
+		} else {
+			volumes = gkr_location_manager_get_volumes (NULL);
+			for (l = volumes; l; l = g_slist_next (l)) {
+				loc = gkr_location_from_child (GPOINTER_TO_UINT (l->data), 
+				                               place->directory);
+				if (loc == parent) {
+					defaults = place->defaults;
+					break;
+				}
+			}
+			g_slist_free (volumes);
+		}
+		
+		/* Found something? */
+		if (defaults)
+			return defaults;
+	}
+	
+	return NULL;	
+}
+
+static GKeyFile*
+load_parent_key_file (GkrPkIndex *index, GQuark loc)
+{
+	GKeyFile *file;
+	GQuark parent;
+	const gchar *defaults;
+	
+	if (!loc)
+		return NULL;
+		
+	parent = gkr_location_to_parent (loc);
+	if (!parent)
+		return NULL;
+		
+	file = g_hash_table_lookup (index->defaults_by_parent, GUINT_TO_POINTER (parent));
+	if (!file) {
+		file = g_key_file_new ();
+		g_hash_table_insert (index->defaults_by_parent, GUINT_TO_POINTER (parent), file);	
+
+		/* 
+		 * Look in the places list and load any default index data 
+		 * from there. 
+		 */
+		defaults = find_parent_defaults (parent);
+		if (defaults) {
+			if (!g_key_file_load_from_data (file, defaults, strlen (defaults), 
+			                                G_KEY_FILE_NONE, NULL))
+				g_warning ("couldn't parse builtin parent defaults");
+		}
+	}
+	
+	return file;
+}
 
 static gboolean
 read_pk_index_value (GkrPkIndex *index, GQuark loc, gkrconstunique uni, 
-                     const gchar *field, GValue *value)
+                     const gchar *field, GkrPkObject *object, 
+                     ReadValueFunc func, gpointer data)
 {
 	const gchar *path = NULL;
 	struct stat sb;
 	gboolean force = FALSE;
 	GKeyFile *key_file = NULL;
+	gchar *group;
+	gint ret = 0;
+	
+	g_return_val_if_fail (uni, FALSE);
 
 	if (loc) {
 		path = index_path_for_location (index, loc);
@@ -452,31 +688,56 @@ read_pk_index_value (GkrPkIndex *index, GQuark loc, gkrconstunique uni,
 	}
 	
 	key_file = load_index_key_file (index, loc, -1, force);
-	if (!key_file)
-		return FALSE;
 	
-	if (!get_keyfile_value (key_file, uni, field, value))
-		return FALSE;
+	/* Try the actual item first */
+	if (key_file) {
+		group = unique_to_group (uni);
+		g_return_val_if_fail (group, FALSE);
 	
-	return TRUE;
+		ret = get_keyfile_value (key_file, group, field, func, data);
+		g_free (group);
+
+		/* If not found, look in the default section */
+		if (ret == 0)
+			ret = get_keyfile_value (key_file, "default", field, func, data);
+	}
+		
+	/* Look in the parent directory defaults */
+	if (ret == 0) {
+		key_file = load_parent_key_file (index, loc);
+		if (key_file) 
+			ret = get_keyfile_value (key_file, "default", field, func, data);
+	}
+
+	/* 
+	 * If we saw that the file was changed, then tell the object
+	 * to flush all of its caches and etc...
+	 */ 
+	if (force && object)
+		gkr_pk_object_flush (object);
+	
+	return ret == 1;
 }
 
 static gboolean
 update_pk_index_value (GkrPkIndex *index, GQuark loc, gkrconstunique uni, 
-                       const gchar *field, GValue *value)
+                       const gchar *field, GkrPkObject *object, 
+                       WriteValueFunc func, gpointer data)
 {
 	const gchar *path = NULL;
-	gchar *data = NULL;
+	gchar *contents = NULL;
 	gboolean ret = FALSE;
 	gboolean force = FALSE;
 	gboolean updated = FALSE;
 	GError *err = NULL;
 	GKeyFile *key_file = NULL;
-	gsize n_data;
+	gsize n_contents;
 	struct stat sb;
 	int tries = 0;
 	int fd = -1;
 	
+	g_return_val_if_fail (uni, FALSE);
+	
 	if (loc) {
 		path = index_path_for_location (index, loc);
 		if (!path) 
@@ -521,12 +782,12 @@ update_pk_index_value (GkrPkIndex *index, GQuark loc, gkrconstunique uni,
 	if (!key_file)
 		goto done;
 
-	set_keyfile_value (key_file, uni, field, value, &updated);
+	set_keyfile_value (key_file, uni, field, func, data, &updated);
 	if (updated && loc) {
 		
 		/* Serialize the key file into memory */
-		data = g_key_file_to_data (key_file, &n_data, &err);
-		if (!data) {
+		contents = g_key_file_to_data (key_file, &n_contents, &err);
+		if (!contents) {
 			g_warning ("couldn't serialize index file: %s", 
 			           err && err->message ? err->message : "");
 			g_error_free (err);
@@ -536,7 +797,7 @@ update_pk_index_value (GkrPkIndex *index, GQuark loc, gkrconstunique uni,
 		g_assert (path);
 		
 		/* And write that memory to disk atomically */
-		if (!g_file_set_contents (path, data, n_data, &err)) {
+		if (!g_file_set_contents (path, contents, n_contents, &err)) {
 			g_message ("couldn't write index file to disk: %s: %s", 
 			           path, err && err->message ? err->message : "");
 			g_error_free (err);
@@ -544,12 +805,19 @@ update_pk_index_value (GkrPkIndex *index, GQuark loc, gkrconstunique uni,
 		}
 	}
 	
+	/* 
+	 * If the file was updated then tell the object to flush all of 
+	 * its caches and other optimizations...
+	 */
+	if ((force || updated) && object)
+		gkr_pk_object_flush (object);
+		
 	ret = TRUE;
 	
 done:
 	if (fd != -1)
 		close (fd);
-	g_free (data);
+	g_free (contents);
 	
 	return ret;	
 }
@@ -595,6 +863,8 @@ flush_caches (GkrLocationManager *locmgr, GQuark volume, GkrPkIndex *index)
 	                             GUINT_TO_POINTER (volume));
 	g_hash_table_foreach_remove (index->mtime_by_location, remove_descendent_locations, 
 	                             GUINT_TO_POINTER (volume));
+	g_hash_table_foreach_remove (index->defaults_by_parent, remove_descendent_locations,
+	                             GUINT_TO_POINTER (volume));
 }
 
 /* -----------------------------------------------------------------------------
@@ -610,6 +880,8 @@ gkr_pk_index_init (GkrPkIndex *index)
 	index->mtime_by_location = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, free_mtime);
 	index->file_by_location = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, 
 	                                                 (GDestroyNotify)g_key_file_free);
+	index->defaults_by_parent = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL,
+	                                                   (GDestroyNotify)g_key_file_free);
 	
 	locmgr = gkr_location_manager_get ();
 	g_signal_connect (locmgr, "volume-removed", G_CALLBACK (flush_caches), index);
@@ -627,7 +899,9 @@ gkr_pk_index_finalize (GObject *obj)
 	g_hash_table_destroy (index->path_by_location);
 	g_hash_table_destroy (index->mtime_by_location);
 	g_hash_table_destroy (index->file_by_location);
-	index->path_by_location = index->mtime_by_location = index->file_by_location = NULL;
+	g_hash_table_destroy (index->defaults_by_parent);
+	index->path_by_location = index->mtime_by_location = NULL;
+	index->file_by_location = index->defaults_by_parent = NULL;
 	
 	G_OBJECT_CLASS (gkr_pk_index_parent_class)->finalize (obj);
 }
@@ -651,77 +925,78 @@ gkr_pk_index_class_init (GkrPkIndexClass *klass)
  */
 
 gboolean
-gkr_pk_index_get_boolean (GQuark loc, gkrconstunique uni, 
-                          const gchar *field, gboolean defvalue)
+gkr_pk_index_get_boolean (GkrPkObject *obj, const gchar *field, gboolean defvalue)
 {
-	GValue value;
 	gboolean ret = defvalue;
 	
-	g_return_val_if_fail (uni != NULL, ret);	
+	g_return_val_if_fail (GKR_IS_PK_OBJECT (obj), ret);
 	g_return_val_if_fail (field != NULL, ret);	
 	
-	memset (&value, 0, sizeof (value));
-	g_value_init (&value, G_TYPE_BOOLEAN);
-	
-	if (read_pk_index_value (get_index_singleton (), loc, uni, field, &value))
-		ret = g_value_get_boolean (&value);
+	if (!read_pk_index_value (get_index_singleton (), obj->location, obj->unique,
+	                          field, obj, (ReadValueFunc)read_boolean_value, &ret))
+		ret = defvalue;
 
-	g_value_unset (&value);	
 	return ret;
 }
 
 gint
-gkr_pk_index_get_int (GQuark loc, gkrconstunique uni, 
-                      const gchar *field, gint defvalue)
+gkr_pk_index_get_int (GkrPkObject *obj, const gchar *field, gint defvalue)
 {
-	GValue value;
 	gint ret = defvalue;
 	
-	g_return_val_if_fail (uni != NULL, ret);	
+	g_return_val_if_fail (GKR_IS_PK_OBJECT (obj), ret);	
 	g_return_val_if_fail (field != NULL, ret);	
-	
-	memset (&value, 0, sizeof (value));
-	g_value_init (&value, G_TYPE_INT);
-	
-	if (read_pk_index_value (get_index_singleton (), loc, uni, field, &value))
-		ret = g_value_get_int (&value);
 
-	g_value_unset (&value);	
+	if (!read_pk_index_value (get_index_singleton (), obj->location, obj->unique,
+	                          field, obj, (ReadValueFunc)read_int_value, &ret))
+		ret = defvalue;
+
 	return ret;	
 }                                                                 
 
 gchar*
-gkr_pk_index_get_string (GQuark loc, gkrconstunique uni, const gchar *field)
+gkr_pk_index_get_string (GkrPkObject *obj, const gchar *field)
 {
-	GValue value;
 	gchar *ret = NULL;
 	
-	g_return_val_if_fail (uni != NULL, NULL);	
+	g_return_val_if_fail (GKR_IS_PK_OBJECT (obj), NULL);
 	g_return_val_if_fail (field != NULL, NULL);	
 	
-	memset (&value, 0, sizeof (value));
-	g_value_init (&value, G_TYPE_STRING);
-	
-	if (read_pk_index_value (get_index_singleton (), loc, uni, field, &value)) {
-		/* No way to steal value's string, so just don't unset it */
-		ret = (gchar*)g_value_get_string (&value);
-	}
-	
+	if (!read_pk_index_value (get_index_singleton (), obj->location, obj->unique,
+	                          field, obj, (ReadValueFunc)read_string_value, &ret))
+		ret = NULL;
+
 	return ret;
 }
 
+gchar*
+gkr_pk_index_get_string_full (GQuark location, gkrconstunique uni, 
+                              const gchar *field)
+{
+	gchar *ret = NULL;
+	
+	g_return_val_if_fail (uni, NULL);
+	g_return_val_if_fail (field != NULL, NULL);	
+	
+	if (!read_pk_index_value (get_index_singleton (), location, uni, field,
+	                          NULL, (ReadValueFunc)read_string_value, &ret))
+		ret = NULL;
+
+	return ret;	
+}
+
 guchar*
-gkr_pk_index_get_binary (GQuark loc, gkrconstunique unique, const gchar *field, gsize *n_data)
+gkr_pk_index_get_binary (GkrPkObject *obj, const gchar *field, gsize *n_data)
 {
 	guchar *data;
 	gchar *str;
 	gsize n_str;
 
-	g_return_val_if_fail (unique != NULL, NULL);	
+	g_return_val_if_fail (GKR_IS_PK_OBJECT (obj), NULL);
 	g_return_val_if_fail (field != NULL, NULL);	
 	g_return_val_if_fail (n_data != NULL, NULL);	
 
-	str = gkr_pk_index_get_string (loc, unique, field);
+	str = gkr_pk_index_get_string (obj, field);
 	if (!str)
 		return NULL;
 		
@@ -738,75 +1013,114 @@ gkr_pk_index_get_binary (GQuark loc, gkrconstunique unique, const gchar *field,
 	return data;	
 }
 
-gboolean
-gkr_pk_index_set_boolean (GQuark loc, gkrconstunique uni, 
-                          const gchar *field, gboolean val)
+GQuark* 
+gkr_pk_index_get_quarks (GkrPkObject *obj, const gchar *field)
 {
-	GValue value;
-	gboolean ret;
-
-	g_return_val_if_fail (uni != NULL, FALSE);	
-	g_return_val_if_fail (field != NULL, FALSE);
+	GQuark *ret = NULL;
+	
+	g_return_val_if_fail (GKR_IS_PK_OBJECT (obj), NULL);
+	g_return_val_if_fail (field != NULL, NULL);	
+	
+	if (!read_pk_index_value (get_index_singleton (), obj->location, obj->unique,
+	                          field, obj, (ReadValueFunc)read_quarks_value, &ret))
+		ret = NULL;
 		
-	memset (&value, 0, sizeof (value));
-	g_value_init (&value, G_TYPE_BOOLEAN);
-	g_value_set_boolean (&value, val);
-	
-	ret = update_pk_index_value (get_index_singleton (), loc, uni, field, &value);
-	g_value_unset (&value);
-	
 	return ret;
 }
 
 gboolean
-gkr_pk_index_set_int (GQuark loc, gkrconstunique uni, 
-                      const gchar *field, gint val)
+gkr_pk_index_has_value (GkrPkObject *obj, const gchar *field)
 {
-	GValue value;
 	gboolean ret;
 
-	g_return_val_if_fail (uni != NULL, FALSE);	
+	g_return_val_if_fail (GKR_IS_PK_OBJECT (obj), FALSE);
+	g_return_val_if_fail (field != NULL, FALSE);
+	
+	if (!read_pk_index_value (get_index_singleton (), obj->location, obj->unique,
+	                          field, obj, (ReadValueFunc)read_exists_value, &ret))
+		ret = FALSE;
+
+	return ret;
+}
+
+gboolean
+gkr_pk_index_have (GkrPkObject *obj)
+{
+	gboolean ret;
+
+	g_return_val_if_fail (GKR_IS_PK_OBJECT (obj), FALSE);
+	
+	if (!read_pk_index_value (get_index_singleton (), obj->location, obj->unique,
+	                          NULL, obj, (ReadValueFunc)read_exists_any_value, &ret))
+		ret = FALSE;
+
+	return ret;
+}
+
+gboolean
+gkr_pk_index_have_full (GQuark location, gkrconstunique uni)
+{
+	gboolean ret;
+
+	g_return_val_if_fail (uni, FALSE);
+	
+	if (!read_pk_index_value (get_index_singleton (), location, uni, NULL,
+	                          NULL, (ReadValueFunc)read_exists_any_value, &ret))
+		ret = FALSE;
+
+	return ret;
+}
+
+gboolean
+gkr_pk_index_set_boolean (GkrPkObject *obj, const gchar *field, gboolean val)
+{
+	g_return_val_if_fail (GKR_IS_PK_OBJECT (obj), FALSE);
 	g_return_val_if_fail (field != NULL, FALSE);
 		
-	memset (&value, 0, sizeof (value));
-	g_value_init (&value, G_TYPE_INT);
-	g_value_set_int (&value, val);
-	
-	ret = update_pk_index_value (get_index_singleton (), loc, uni, field, &value);
-	g_value_unset (&value);
-	
-	return ret;
+	return update_pk_index_value (get_index_singleton (), obj->location, obj->unique, 
+	                              field, obj, (WriteValueFunc)write_boolean_value, &val);
+}
+
+gboolean
+gkr_pk_index_set_int (GkrPkObject *obj, const gchar *field, gint val)
+{
+	g_return_val_if_fail (GKR_IS_PK_OBJECT (obj), FALSE);
+	g_return_val_if_fail (field != NULL, FALSE);
+
+	return update_pk_index_value (get_index_singleton (), obj->location, obj->unique, 
+	                              field, obj, (WriteValueFunc)write_int_value, &val);
 }                                                       
                                                         
 gboolean 
-gkr_pk_index_set_string (GQuark loc, gkrconstunique uni, const gchar *field, const gchar *val)
+gkr_pk_index_set_string (GkrPkObject *obj, const gchar *field, const gchar *val)
 {
-	GValue value;
-	gboolean ret;
-
-	g_return_val_if_fail (uni != NULL, FALSE);	
+	g_return_val_if_fail (GKR_IS_PK_OBJECT (obj), FALSE);
 	g_return_val_if_fail (field != NULL, FALSE);
-	g_return_val_if_fail (val != NULL, FALSE);
-		
-	memset (&value, 0, sizeof (value));
-	g_value_init (&value, G_TYPE_STRING);
-	g_value_set_string (&value, val);
-	
-	ret = update_pk_index_value (get_index_singleton (), loc, uni, field, &value);
-	g_value_unset (&value);
-	
-	return ret;
+
+	return update_pk_index_value (get_index_singleton (), obj->location, obj->unique, 
+	                              field, obj, (WriteValueFunc)write_string_value, &val);
 }
 
 gboolean
-gkr_pk_index_set_binary (GQuark loc, gkrconstunique unique, const gchar *field, 
+gkr_pk_index_set_string_full (GQuark location, gkrconstunique uni, const gchar *field, 
+                              const gchar *val)
+{
+	g_return_val_if_fail (uni, FALSE);
+	g_return_val_if_fail (field != NULL, FALSE);
+
+	return update_pk_index_value (get_index_singleton (), location, uni, field, 
+	                              NULL, (WriteValueFunc)write_string_value, &val);	
+}
+
+gboolean
+gkr_pk_index_set_binary (GkrPkObject *obj, const gchar *field, 
                          const guchar *data, gsize n_data)
 {
 	gboolean ret, r;
 	gchar *str;
 	gsize n_str;
 	
-	g_return_val_if_fail (unique != NULL, FALSE);	
+	g_return_val_if_fail (GKR_IS_PK_OBJECT (obj), FALSE);
 	g_return_val_if_fail (field != NULL, FALSE);
 	g_return_val_if_fail (data != NULL, FALSE);
 	
@@ -816,28 +1130,71 @@ gkr_pk_index_set_binary (GQuark loc, gkrconstunique unique, const gchar *field,
 	r = gkr_crypto_hex_encode (data, n_data, str, &n_str);
 	g_assert (r == TRUE);
 	
-	ret = gkr_pk_index_set_string (loc, unique, field, str);
+	ret = gkr_pk_index_set_string (obj, field, str);
 	g_free (str);
 
 	return ret;
 }
 
 gboolean
-gkr_pk_index_delete (GQuark loc, gkrconstunique uni, const gchar *field)
+gkr_pk_index_set_quarks (GkrPkObject *obj, const gchar *field, GQuark *quarks)
 {
-	GValue value;
-	gboolean ret;
-
-	g_return_val_if_fail (uni != NULL, FALSE);	
+	g_return_val_if_fail (GKR_IS_PK_OBJECT (obj), FALSE);
 	g_return_val_if_fail (field != NULL, FALSE);
-	
-	/* Values can't be set to NULL, so for us an empty pointer means delete */
-	memset (&value, 0, sizeof (value));
-	g_value_init (&value, G_TYPE_POINTER);
-	g_value_set_pointer (&value, NULL);
-	
-	ret = update_pk_index_value (get_index_singleton (), loc, uni, field, &value);
-	g_value_unset (&value);
 
-	return ret;	
+	return update_pk_index_value (get_index_singleton (), obj->location, obj->unique, 
+	                              field, obj, (WriteValueFunc)write_quarks_value, &quarks);
+}
+
+gboolean
+gkr_pk_index_delete (GkrPkObject *obj, const gchar *field)
+{
+	g_return_val_if_fail (GKR_IS_PK_OBJECT (obj), FALSE);
+	g_return_val_if_fail (field != NULL, FALSE);
+
+	return update_pk_index_value (get_index_singleton (), obj->location, obj->unique, 
+	                              field, obj, (WriteValueFunc)write_delete, NULL);
+
+}
+
+gboolean
+gkr_pk_index_clear (GkrPkObject *obj)
+{
+	g_return_val_if_fail (GKR_IS_PK_OBJECT (obj), FALSE);
+
+	return update_pk_index_value (get_index_singleton (), obj->location, obj->unique, 
+	                              NULL, obj, (WriteValueFunc)write_clear, NULL);
+
+}
+
+gboolean
+gkr_pk_index_quarks_has (GQuark *quarks, GQuark check)
+{
+	while (*quarks) {
+		if (*quarks == check)
+			return TRUE;
+		++quarks;
+	}
+	
+	return FALSE;
+}
+
+GQuark*
+gkr_pk_index_quarks_dup (GQuark *quarks)
+{
+	GQuark *last = quarks;
+	
+	/* Figure out how many there are */	
+	while (*last)
+		++last;
+		
+	/* Include the null termination */
+	++last;
+	return g_memdup (quarks, (last - quarks) * sizeof (GQuark));
+}
+
+void
+gkr_pk_index_quarks_free (GQuark *quarks)
+{
+	g_free (quarks);
 }
